import numpy as np
import vg


def rotation_from_up_and_look(up, look):
    """
    Rotation matrix to rotate a mesh into a canonical reference frame. The
    result is a rotation matrix that will make up along +y and look along +z
    (i.e. facing towards a default opengl camera).

    up: The direction you want to become `+y`.
    look: The direction you want to become `+z`.

    """
    vg.shape.check(locals(), "up", (3,))
    vg.shape.check(locals(), "look", (3,))

    up, look = [np.asarray(vector, dtype=np.float64) for vector in (up, look)]

    if np.linalg.norm(up) == 0:
        raise ValueError("Singular up")
    if np.linalg.norm(look) == 0:
        raise ValueError("Singular look")

    y = up / np.linalg.norm(up)
    z = look - np.dot(look, y) * y
    if np.linalg.norm(z) == 0:
        raise ValueError("up and look are collinear")
    z = z / np.linalg.norm(z)
    x = np.cross(y, z)
    return np.array([x, y, z])


def euler(xyz, order="xyz", units="deg"):
    """
    Convert a Euler angle representation of 3D rotations to a 3x3 rotation matrix.

    Euler angles are a way of representing 3D rotations as a sequence of rotations
    about the axes. Conceptually, think of `euler([10, 20, 30])` as
    "Rotate 10 degrees around the x axis, then 20 degrees around the y axis, then
    30 degrees around the z axis" (that ordering can be changed with the `order`
    argument, and the units can be given in degrees or radians by setting `units`
    to `'deg'` or `'rad'`).

    Euler angles are a problematic representation of rotation for numerical methods,
    as there are multiple possible representations for a given rotation. But they are
    a very intuitive and readable way to initialize a rotation matrix.

    See also:

        - https://en.wikipedia.org/wiki/Euler_angles
    """
    if not hasattr(xyz, "__iter__"):
        xyz = [xyz]
    if units == "deg":
        xyz = np.radians(xyz)
    r = np.eye(3)
    for theta, axis in zip(xyz, order):
        c = np.cos(theta)
        s = np.sin(theta)
        if axis == "x":
            r = np.dot(np.array([[1, 0, 0], [0, c, -s], [0, s, c]]), r)
        if axis == "y":
            r = np.dot(np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]]), r)
        if axis == "z":
            r = np.dot(np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]]), r)
    return r



import numpy as np

__all__ = [
    "create_rectangular_prism",
    "create_cube",
    "create_triangular_prism",
    "create_rectangle",
]


def _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces):
    if ret_unique_vertices_and_faces:
        return vertices, faces
    else:
        return vertices[faces]


def create_rectangular_prism(origin, size, ret_unique_vertices_and_faces=False):
    """
    Tesselate an axis-aligned rectangular prism. One vertex is `origin`. The
    diametrically opposite vertex is `origin + size`.

    Args:
        origin (np.ndarray): A 3D point vector containing the point on the
            prism with the minimum x, y, and z coords.
        size (np.ndarray): A 3D vector specifying the prism's length, width,
            and height, which should be positive.
        ret_unique_vertices_and_faces (bool): When `True` return a vertex
            array containing the unique vertices and an array of faces (i.e.
            vertex indices). When `False`, return a flattened array of
            triangle coordinates.

    Returns:
        object:

        - With `ret_unique_vertices_and_faces=True`: a tuple containing
          an `8x3` array of vertices and a `12x3` array of triangle faces.
        - With `ret_unique_vertices_and_faces=False`: a `12x3x3` matrix of
          flattened triangle coordinates.
    """
    from ..tri import quads_to_tris

    lower_base_plane = np.array(
        [
            # Lower base plane
            origin,
            origin + np.array([size[0], 0, 0]),
            origin + np.array([size[0], 0, size[2]]),
            origin + np.array([0, 0, size[2]]),
        ]
    )
    upper_base_plane = lower_base_plane + np.array([0, size[1], 0])

    vertices = np.vstack([lower_base_plane, upper_base_plane])

    faces = np.array(
        quads_to_tris(
            np.array(
                [
                    [0, 1, 2, 3],  # lower base (-y)
                    [7, 6, 5, 4],  # upper base (+y)
                    [4, 5, 1, 0],  # +z face
                    [5, 6, 2, 1],  # +x face
                    [6, 7, 3, 2],  # -z face
                    [3, 7, 4, 0],  # -x face
                ],
                dtype=np.uint64,
            )
        ),
        dtype=np.uint64,
    )

    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)


def create_cube(origin, size, ret_unique_vertices_and_faces=False):
    """
    Tesselate an axis-aligned cube. One vertex is `origin`. The diametrically
    opposite vertex is `size` units along `+x`, `+y`, and `+z`.

    Args:
        origin (np.ndarray): A 3D point vector containing the point on the
            prism with the minimum x, y, and z coords.
        size (float): The length, width, and height of the cube, which should
            be positive.
        ret_unique_vertices_and_faces (bool): When `True` return a vertex
            array containing the unique vertices and an array of faces (i.e.
            vertex indices). When `False`, return a flattened array of
            triangle coordinates.

    Returns:
        object:

        - With `ret_unique_vertices_and_faces=True`: a tuple containing
          an `8x3` array of vertices and a `12x3` array of triangle faces.
        - With `ret_unique_vertices_and_faces=False`: a `12x3x3` matrix of
          flattened triangle coordinates.
    """
    return create_rectangular_prism(
        origin,
        np.repeat(size, 3),
        ret_unique_vertices_and_faces=ret_unique_vertices_and_faces,
    )


def create_triangular_prism(p1, p2, p3, height, ret_unique_vertices_and_faces=False):
    """
    Tesselate a triangular prism whose base is the triangle `p1`, `p2`, `p3`.
    If the vertices are oriented in a counterclockwise direction, the prism
    extends from behind them.

    Args:
        p1 (np.ndarray): A 3D point on the base of the prism.
        p2 (np.ndarray): A 3D point on the base of the prism.
        p3 (np.ndarray): A 3D point on the base of the prism.
        height (float): The height of the prism, which should be positive.
        ret_unique_vertices_and_faces (bool): When `True` return a vertex
            array containing the unique vertices and an array of faces (i.e.
            vertex indices). When `False`, return a flattened array of
            triangle coordinates.

    Returns:
        object:

        - With `ret_unique_vertices_and_faces=True`: a tuple containing
          an `6x3` array of vertices and a `8x3` array of triangle faces.
        - With `ret_unique_vertices_and_faces=False`: a `8x3x3` matrix of
          flattened triangle coordinates.
    """
    from .. import Plane

    base_plane = Plane.from_points(p1, p2, p3)
    lower_base_to_upper_base = height * -base_plane.normal
    vertices = np.vstack(([p1, p2, p3], [p1, p2, p3] + lower_base_to_upper_base))

    faces = np.array(
        [
            [0, 1, 2],  # base
            [0, 3, 4],
            [0, 4, 1],  # side 0, 3, 4, 1
            [1, 4, 5],
            [1, 5, 2],  # side 1, 4, 5, 2
            [2, 5, 3],
            [2, 3, 0],  # side 2, 5, 3, 0
            [5, 4, 3],  # base
        ],
        dtype=np.uint64,
    )

    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)


def create_rectangle(ret_unique_vertices_and_faces=False):
    """
    Create a rectangle.

    Args:
        ret_unique_vertices_and_faces (bool): When `True` return a vertex
            array containing the unique vertices and an array of faces (i.e.
            vertex indices). When `False`, return a flattened array of
            triangle coordinates.

    Returns:
        object:

        - With `ret_unique_vertices_and_faces=True`: a tuple containing
          an `4x3` array of vertices and a `2x3` array of triangle faces.
        - With `ret_unique_vertices_and_faces=False`: a `16x3x3` matrix of
          flattened triangle coordinates.
    """
    vertices = np.array(
        [[1.0, 0.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, -1.0]]
    )
    faces = np.array([[0, 1, 2], [3, 1, 0]], dtype=np.uint64)
    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)



import numpy as np
import vg


def rodrigues_vector_to_rotation_matrix(r, calculate_jacobian=False):
    """
    Convert a 3x1 or 1x3 Rodrigues vector to a 3x3 rotation matrix.

    A Rodrigues vector is a 3 element vector representing a 3D rotation.
    Its direction represents the axis about which to rotate and its magnitude
    represents the amount to rotate by.

    All of SO3 (that is, all 3D rotations) can be uniquely represented by a
    Rodrigues vector, and it does not suffer from the multiple representation
    and gimbal locking problems that Euler angle representations do.

    If `calculate_jacobian` is passed, then the derivative of the rotation is
    also computed. Note that the derivative is undefined for a Rodrigues vector
    of `[0,0,0]` (that is, no rotation).

    See also:

        - https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    r = np.array(r, dtype=np.double)
    r = r.flatten()
    vg.shape.check_value(r, (3,))
    eps = np.finfo(np.double).eps
    theta = np.linalg.norm(r)
    if theta < eps:
        r_out = np.eye(3)
        if calculate_jacobian:
            jac = np.zeros((3, 9))
            jac[0, 5] = jac[1, 6] = jac[2, 1] = -1
            jac[0, 7] = jac[1, 2] = jac[2, 3] = 1
    else:
        c = np.cos(theta)
        s = np.sin(theta)
        c1 = 1.0 - c
        itheta = 1.0 if theta == 0.0 else 1.0 / theta
        r *= itheta
        I = np.eye(3)  # noqa: E741 I is an identity matrix
        rrt = np.array([r * r[0], r * r[1], r * r[2]])
        _r_x_ = np.array([[0, -r[2], r[1]], [r[2], 0, -r[0]], [-r[1], r[0], 0]])
        r_out = c * I + c1 * rrt + s * _r_x_
        if calculate_jacobian:
            drrt = np.array(
                [
                    [r[0] + r[0], r[1], r[2], r[1], 0, 0, r[2], 0, 0],
                    [0, r[0], 0, r[0], r[1] + r[1], r[2], 0, r[2], 0],
                    [0, 0, r[0], 0, 0, r[1], r[0], r[1], r[2] + r[2]],
                ]
            )
            d_r_x_ = np.array(
                [
                    [0, 0, 0, 0, 0, -1, 0, 1, 0],
                    [0, 0, 1, 0, 0, 0, -1, 0, 0],
                    [0, -1, 0, 1, 0, 0, 0, 0, 0],
                ]
            )
            I_jac = np.array([I.flatten(), I.flatten(), I.flatten()])
            ri = np.array([[r[0]], [r[1]], [r[2]]])
            a0 = -s * ri
            a1 = (s - 2 * c1 * itheta) * ri
            a2 = np.ones((3, 1)) * c1 * itheta
            a3 = (c - s * itheta) * ri
            a4 = np.ones((3, 1)) * s * itheta
            jac = (
                a0 * I_jac
                + a1 * rrt.flatten()
                + a2 * drrt
                + a3 * _r_x_.flatten()
                + a4 * d_r_x_
            )
    if calculate_jacobian:
        return r_out, jac
    else:
        return r_out


def rotation_matrix_to_rodrigues_vector(r, calculate_jacobian=False):
    """
    Convert a 3x3 rotation matrix to a 3x1 or 1x3 Rodrigues vector.

    A Rodrigues vector is a 3 element vector representing a 3D rotation.
    Its direction represents the axis about which to rotate and its magnitude
    represents the amount to rotate by.

    All of SO3 (that is, all 3D rotations) can be uniquely represented by a
    Rodrigues vector, and it does not suffer from the multiple representation
    and gimbal locking problems that Euler angle representations do.

    If `calculate_jacobian` is passed, then the derivative of the rotation is
    also computed. Note that the derivative is undefined for a Rodrigues vector
    of `[0,0,0]` (that is, no rotation).

    See also:

        - https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    r = np.array(r, dtype=np.double)
    vg.shape.check_value(r, (3, 3))
    u, _, v = np.linalg.svd(r)
    r = np.dot(u, v)
    rx = r[2, 1] - r[1, 2]
    ry = r[0, 2] - r[2, 0]
    rz = r[1, 0] - r[0, 1]
    s = np.linalg.norm(np.array([rx, ry, rz])) * np.sqrt(0.25)
    c = np.clip((np.sum(np.diag(r)) - 1) * 0.5, -1, 1)
    theta = np.arccos(c)
    if s < 1e-5:
        if c > 0:
            r_out = np.zeros((3, 1))
        else:
            rx, ry, rz = np.clip(np.sqrt((np.diag(r) + 1) * 0.5), 0, np.inf)
            if r[0, 1] < 0:
                ry = -ry
            if r[0, 2] < 0:
                rz = -rz
            if (
                np.abs(rx) < np.abs(ry)
                and np.abs(rx) < np.abs(rz)
                and ((r[1, 2] > 0) != (ry * rz > 0))
            ):
                rz = -rz
            r_out = np.array([[rx, ry, rz]]).T
            theta /= np.linalg.norm(r_out)
            r_out *= theta
        if calculate_jacobian:
            jac = np.zeros((9, 3))
            if c > 0:
                jac[1, 2] = jac[5, 0] = jac[6, 1] = -0.5
                jac[2, 1] = jac[3, 2] = jac[7, 0] = 0.5
    else:
        vth = 1.0 / (2.0 * s)
        if calculate_jacobian:
            dtheta_dtr = -1.0 / s
            dvth_dtheta = -vth * c / s
            d1 = 0.5 * dvth_dtheta * dtheta_dtr
            d2 = 0.5 * dtheta_dtr
            dvardR = np.array(
                [
                    [0, 0, 0, 0, 0, 1, 0, -1, 0],
                    [0, 0, -1, 0, 0, 0, 1, 0, 0],
                    [0, 1, 0, -1, 0, 0, 0, 0, 0],
                    [d1, 0, 0, 0, d1, 0, 0, 0, d1],
                    [d2, 0, 0, 0, d2, 0, 0, 0, d2],
                ]
            )
            dvar2dvar = np.array(
                [
                    [vth, 0, 0, rx, 0],
                    [0, vth, 0, ry, 0],
                    [0, 0, vth, rz, 0],
                    [0, 0, 0, 0, 1],
                ]
            )
            domegadvar2 = np.array(
                [
                    [theta, 0, 0, rx * vth],
                    [0, theta, 0, ry * vth],
                    [0, 0, theta, rz * vth],
                ]
            )
            jac = np.dot(np.dot(domegadvar2, dvar2dvar), dvardR)
            for ii in range(3):
                jac[ii] = jac[ii].reshape((3, 3)).T.flatten()
            jac = jac.T
        vth *= theta
        r_out = np.array([[rx, ry, rz]]).T * vth
    if calculate_jacobian:
        return r_out, jac
    else:
        return r_out


def cv2_rodrigues(r, calculate_jacobian=False):
    """
    `cv2_rodrigues` is a wrapped function designed to be API compatible with
    OpenCV's `cv2.Rodrigues`.

    If it is given a rotation matrix, it returns a Rodrigues vector.

    If it is given a Rodrigues vector, it returns a rotation matrix.

    To make your code clearer, call `rodrigues_vector_to_rotation_matrix` or
    `rotation_matrix_to_rodrigues_vector` directly, which makes the intent of
    your code clearer.
    """
    r = np.array(r, dtype=np.double)
    if r.size == 3:
        return rodrigues_vector_to_rotation_matrix(r, calculate_jacobian)
    elif r.shape == (3, 3):
        return rotation_matrix_to_rodrigues_vector(r, calculate_jacobian)
    else:
        raise ValueError("rodrigues: input matrix must be 1x3, 3x1 or 3x3.")



from ._segment_functions import (
    closest_point_of_line_segment,
    subdivide_segment,
    subdivide_segments,
)

__all__ = ["closest_point_of_line_segment", "subdivide_segment", "subdivide_segments"]



import numpy as np
import vg


class Box(object):
    """
    An axis-aligned cuboid or rectangular prism. It's defined by an origin
    point, which is its minimum point in each dimension, and non-negative size
    (length, width, and depth).

    Args:
        origin (np.arraylike): The `x`, `y`, and `z` coordinate of the
            origin, the minimum point in each dimension.
        size (np.arraylike): An array containing the width (dx), height
            (dy), and depth (dz), which must be non-negative.
    """

    def __init__(self, origin, size):
        vg.shape.check(locals(), "origin", (3,))
        vg.shape.check(locals(), "size", (3,))
        if any(np.less(size, 0)):
            raise ValueError("Shape should be zero or positive")
        self.origin = origin
        self.size = size

    @classmethod
    def from_points(cls, points):
        """
        The smallest box which spans the given points.

        Args:
            points (np.arraylike): A `kx3` array of points.

        Returns:
            Box: The smallest box which spans the given points.
        """
        k = vg.shape.check(locals(), "points", (-1, 3))
        if k == 0:
            raise ValueError("Need at least 1 point")
        return cls(np.min(points, axis=0), np.ptp(points, axis=0))

    @property
    def ranges(self):
        """
        Ranges for each coordinate axis as a 3x2 `np.ndarray`.
        """
        ranges = np.array([self.origin, self.origin + self.size]).T
        # ranges is almost, but not quite what we want, since it might
        # include mins which are greater than maxes, and vice versa.
        # TODO: Is this really true? `self.size` is nonnegative...
        return np.vstack([ranges.min(axis=1), ranges.max(axis=1)]).T

    @property
    def min_x(self):
        """
        The box's minimum `x` coordinate.
        """
        return self.origin[0]

    @property
    def min_y(self):
        """
        The box's minimum `y` coordinate.
        """
        return self.origin[1]

    @property
    def min_z(self):
        """
        The box's minimum `z` coordinate.
        """
        return self.origin[2]

    @property
    def max_x(self):
        """
        The box's maximum `x` coordinate.
        """
        return self.origin[0] + self.size[0]

    @property
    def max_y(self):
        """
        The box's maximum `y` coordinate.
        """
        return self.origin[1] + self.size[1]

    @property
    def max_z(self):
        """
        The box's maximum `z` coordinate.
        """
        return self.origin[2] + self.size[2]

    @property
    def mid_x(self):
        """
        The `x` coordinate of the box's center.
        """
        return self.origin[0] + self.size[0] / 2

    @property
    def mid_y(self):
        """
        The `y` coordinate of the box's center.
        """
        return self.origin[1] + self.size[1] / 2

    @property
    def mid_z(self):
        """
        The `z` coordinate of the box's center.
        """
        return self.origin[2] + self.size[2] / 2

    @property
    def width(self):
        """
        The box's width. Same as `max_x - min_x`.
        """
        return self.size[0]

    @property
    def height(self):
        """
        The box's height. Same as `max_y - min_y`.
        """
        return self.size[1]

    @property
    def depth(self):
        """
        The box's depth. Same as `max_z - min_z`.
        """
        return self.size[2]

    @property
    def center_point(self):
        """
        The box's geometric center.
        """
        return self.origin + 0.5 * self.size

    @property
    def floor_point(self):
        """
        The center of the side of the box having the minimum `y` coordinate.
        This is `center_point` projected to the the level of `min_y`.
        """
        return self.origin + [0.5, 0.0, 0.5] * self.size

    @property
    def volume(self):
        """
        The box's volume.
        """
        return np.prod(self.size)

    @property
    def surface_area(self):
        """
        The box's surface area.
        """
        l, h, w = self.size
        return 2 * (w * l + h * l + h * w)

    @property
    def v(self):
        """
        Corners of the box as an `8x3` array of coordinates.
        """
        return np.array(
            [
                self.origin,
                self.origin + np.array([self.size[0], 0, 0]),
                self.origin + np.array([0, self.size[1], 0]),
                self.origin + np.array([0, 0, self.size[2]]),
                self.origin + np.array([self.size[0], self.size[1], 0]),
                self.origin + np.array([0, self.size[1], self.size[2]]),
                self.origin + np.array([self.size[0], 0, self.size[2]]),
                self.origin + np.array([self.size[0], self.size[1], self.size[2]]),
            ]
        )

    def contains(self, point, atol=None):
        """
        Test whether the box contains the given point. When `atol` is
        provided, returns `True` for points inside the box and points
        whose coordinates are all within `atol` of the box boundary.
        """
        vg.shape.check(locals(), "point", (3,))

        if atol is None:
            atol = 0.0
        return np.all(
            np.logical_and(
                self.origin - atol <= point, point <= self.origin + self.size + atol
            )
        )



import numpy as np


def find_repeats(arr, wrap=False):
    """
    Find the indices of the elements in an array which are equal to the
    elements which immediately precede them.

    To make it easy to work with open and closed polygonal chains, the shape
    of the output will always have the same length as the shape of the input.

    When `wrap` is `True` the first value will be compared to the last value.
    When `wrap` is `False` the first index of the result will always be False.
    """
    if wrap:
        return np.roll(arr, 1) == arr
    else:
        return np.concatenate(([False], arr[:-1] == arr[1:]))


def find_changes(arr, wrap=False):
    """
    Find the indices of the elements in an array which differ from the
    elements which immediately precede them.

    To make it easy to work with open and closed polygonal chains, the shape
    of the output will always have the same length as the shape of the input.

    When `wrap` is `True` the first value will be compared to the last value.
    When `wrap` is `False` the first index of the result will always be False.
    """
    if wrap:
        return np.roll(arr, 1) != arr
    else:
        return np.concatenate(([False], arr[:-1] != arr[1:]))



import numpy as np
import vg
from .._common.shape import columnize

__all__ = [
    "apply_affine_transform",
    "transform_matrix_for_rotation",
    "transform_matrix_for_translation",
    "transform_matrix_for_scale",
]


def apply_affine_transform(points, transform_matrix):
    """
    Apply the given transformation matrix to the points using homogenous
    coordinates.

    Note:
        This works on any transformation matrix, whether or not it is affine.
    """
    vg.shape.check(locals(), "transform_matrix", (4, 4))
    points, _, maybe_decolumnize = columnize(points, (-1, 3), name="points")

    padded_points = np.pad(points, ((0, 0), (0, 1)), mode="constant", constant_values=1)
    transformed_padded_points = np.dot(transform_matrix, padded_points.T).T
    transformed_points = np.delete(transformed_padded_points, 3, axis=1)

    return maybe_decolumnize(transformed_points)


def _convert_33_to_44(matrix):
    """
    Transform from:
        array([[1., 2., 3.],
               [2., 3., 4.],
               [5., 6., 7.]])
    to:
        array([[1., 2., 3., 0.],
               [2., 3., 4., 0.],
               [5., 6., 7., 0.],
               [0., 0., 0., 1.]])

    """
    vg.shape.check(locals(), "matrix", (3, 3))
    result = np.pad(matrix, ((0, 1), (0, 1)), mode="constant")
    result[3][3] = 1
    return result


def transform_matrix_for_rotation(rotation, ret_inverse_matrix=False):
    """
    Create a transformation matrix from the given 3x3 rotation matrix or a
    Rodrigues vector.

    With `ret_inverse_matrix=True`, also returns a matrix which provides
    the reverse transform.
    """
    from ._rodrigues import rodrigues_vector_to_rotation_matrix

    if rotation.shape == (3, 3):
        forward3 = rotation
    else:
        vg.shape.check(locals(), "rotation", (3,))
        forward3 = rodrigues_vector_to_rotation_matrix(rotation)

    forward = _convert_33_to_44(forward3)

    if not ret_inverse_matrix:
        return forward

    # The inverse of a rotation matrix is its transpose.
    inverse = forward.T
    return forward, inverse


def transform_matrix_for_translation(translation, ret_inverse_matrix=False):
    """
    Create a transformation matrix which translates by the provided
    displacement vector.

    Forward:

        [[  1,  0,  0,  v_0 ],
        [  0,  1,  0,  v_1 ],
        [  0,  0,  1,  v_2 ],
        [  0,  0,  0,  1   ]]

    Reverse:

        [[  1,  0,  0,  -v_0 ],
        [  0,  1,  0,  -v_1 ],
        [  0,  0,  1,  -v_2 ],
        [  0,  0,  0,  1    ]]

    Args:
        vector (np.arraylike): A 3x1 vector.
    """
    vg.shape.check(locals(), "translation", (3,))

    forward = np.eye(4)
    forward[:, -1][:-1] = translation

    if not ret_inverse_matrix:
        return forward

    inverse = np.eye(4)
    inverse[:, -1][:-1] = -translation
    return forward, inverse


def transform_matrix_for_scale(scale_factor, ret_inverse_matrix=False):
    """
    Create a transformation matrix that scales by the given factor.

    Forward:
        [[  s_0, 0,   0,   0 ],
        [  0,   s_1, 0,   0 ],
        [  0,   0,   s_2, 0 ],
        [  0,   0,   0,   1 ]]

    Reverse:
        [[  1/s_0, 0,     0,     0 ],
        [  0,     1/s_1, 0,     0 ],
        [  0,     0,     1/s_2, 0 ],
        [  0,     0,     0,     1 ]]

    Args:
        factor (float): The scale factor.
        ret_inverse_matrix (bool): When `True`, also returns a matrix which
            provides the inverse transform.
    """
    if scale_factor <= 0:
        raise ValueError("Scale factor should be greater than zero")

    forward = _convert_33_to_44(np.diag(np.repeat(scale_factor, 3)))

    if not ret_inverse_matrix:
        return forward

    inverse = _convert_33_to_44(np.diag(np.repeat(1.0 / scale_factor, 3)))
    return forward, inverse



from . import _affine_transform
from ._affine_transform import *  # noqa: F401,F403
from ._rodrigues import (  # noqa: F401
    cv2_rodrigues,
    rodrigues_vector_to_rotation_matrix,
    rotation_matrix_to_rodrigues_vector,
)
from ._rotation import euler, rotation_from_up_and_look  # noqa: F401

__all__ = [
    "euler",
    "rodrigues_vector_to_rotation_matrix",
    "rotation_matrix_to_rodrigues_vector",
    "cv2_rodrigues",
    "rotation_from_up_and_look",
] + _affine_transform.__all__



import vg


def columnize(arr, shape=(-1, 3), name=None):
    """
    Helper for functions which may accept many stacks of three points (kx3)
    returning a stack of results, or a single set of three points (3x1)
    returning a single result.

    Returns the points as kx3, and a `transform_result` function which can
    be applied to the result. It picks off the first result in the 3x1 case.

    Not limited to kx3; this can be used different dimensional shapes like
    kx4, or higher dimensional shapes like kx3x3.
    """
    if not isinstance(shape, tuple):
        raise ValueError("shape should be a tuple")
    name = name or "arr"

    if arr.ndim == len(shape):
        vg.shape.check_value(arr, shape, name=name)
        return arr, True, lambda x: x
    else:
        vg.shape.check_value(arr, shape[1:], name=name)
        return arr.reshape(*shape), False, lambda x: x[0]


def check_shape_any(arr, *shapes, name=None):
    if len(shapes) == 0:
        raise ValueError("At least one shape is required")
    for shape in shapes:
        try:
            return vg.shape.check_value(arr, shape, name=name)
        except ValueError:
            pass

    if name is None:
        preamble = "Expected an array"
    else:
        preamble = "Expected {} to be an array".format(name)

    shape_choices = ", ".join(
        shapes[:-2] + (" or ".join([str(shapes[-2]), str(shapes[-1])]),)
    )

    if arr is None:
        raise ValueError("{} with shape {}; got None".format(preamble, shape_choices))
    else:
        raise ValueError(
            "{} with shape {}; got {}".format(preamble, shape_choices, arr.shape)
        )



#!/usr/bin/env python

import os
import sys
import click
import numpy as np

try:
    import cv2
except ImportError:
    print(
        "Error: This script requires opencv, "
        "which is not a runtime or dev dependency "
        "and needs to be installed separately."
    )
    exit(-1)
sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from polliwog.transform.rotation import (
    euler,
)  # noqa: E402 Needs to happen after the modification of sys.path


THINGS_TO_TEST = [
    np.eye(3),
    np.array([0.0, 0.0, 0.0]),
    np.array([[0.0, 0.0, 0.0]]),
    np.array([[0.0, 0.0, 0.0]]).T,
    np.array([0.5, 0.0, 0.0]),
    np.array([1.0, 0.0, 0.0]),
    np.array([10.0, 11.0, 12.0]),  # make sure we work for rotations > 2pi
    euler([45, 0, 0]),  # these are rotation matrices
    euler([0, 32, 0]),
    euler([0, 0, 67]),
    euler([45, 37, 19]),
    np.array(
        [0.644164, 0.726324, 0.051697]
    ),  # approx the rodrigues of the previous matrix
    # test special cases:
    np.array([[0.0, 1.0, 2.0], [1.0, 0.0, 3.0], [2.0, 3.0, 0.0]]),
    np.array([[0.0, -1.0, 2.0], [-1.0, 0.0, 3.0], [2.0, 3.0, 0.0]]),
    np.array([[0.0, 1.0, -2.0], [1.0, 0.0, 3.0], [-2.0, 3.0, 0.0]]),
    np.array([[0.1, 1.0, 2.0], [1.0, 1.7, -3.0], [2.0, -3.0, 2.8]]),
]


@click.command()
def main():
    print("Building Rodrigues test assets")
    print(f"OpenCV Version: {cv2.__version__}")
    test_assets = [(x, cv2.Rodrigues(x)) for x in THINGS_TO_TEST]
    print(f"Produced {len(test_assets)} test assets")
    test_data_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "rodrigues_test_assets.npz"
    )
    print(f"Writing to {test_data_path}")
    with open(test_data_path, "wb") as f:
        np.savez(f, version=cv2.__version__, test_assets=test_assets)
    print(f"Checking {test_data_path}")
    with open(test_data_path, "rb") as f:
        data = np.load(f, allow_pickle=True)
        print(f"Found opencv version {data['version']}")
        print(f"Found {len(data['test_assets'])} assets")


if __name__ == "__main__":
    main()



import numpy as np
import vg
from ._plane_functions import (
    plane_normal_from_points,
    project_point_to_plane,
    signed_distance_to_plane,
)


class Plane(object):
    """
    A 2-D plane in 3-space (not a hyperplane).

    Args:
        point_on_plane (np.arraylike): A reference point on the plane, as a
            NumPy array with three coordinates.
        unit_normal (np.arraylike): The plane normal vector, as a NumPy
            array with three coordinates.
    """

    def __init__(self, point_on_plane, unit_normal):
        vg.shape.check(locals(), "point_on_plane", (3,))
        vg.shape.check(locals(), "unit_normal", (3,))

        if vg.almost_zero(unit_normal):
            raise ValueError("unit_normal should not be the zero vector")

        unit_normal = vg.normalize(unit_normal)

        self._r0 = np.asarray(point_on_plane)
        self._n = np.asarray(unit_normal)

    def __repr__(self):
        return "<Plane of {} through {}>".format(self.normal, self.reference_point)

    @classmethod
    def from_points(cls, p1, p2, p3):
        """
        If the points are oriented in a counterclockwise direction, the plane's
        normal extends towards you.

        """
        vg.shape.check(locals(), "p1", (3,))
        vg.shape.check(locals(), "p2", (3,))
        vg.shape.check(locals(), "p3", (3,))
        points = np.array([p1, p2, p3])
        return cls(point_on_plane=p1, unit_normal=plane_normal_from_points(points))

    @classmethod
    def from_points_and_vector(cls, p1, p2, vector):
        """
        Compute a plane which contains two given points and the given
        vector. Its reference point will be p1.

        For example, to find the vertical plane that passes through
        two landmarks:

            from_points_and_normal(p1, p2, vector)

        Another way to think about this: identify the plane to which
        your result plane should be perpendicular, and specify vector
        as its normal vector.

        """
        vg.shape.check(locals(), "p1", (3,))
        vg.shape.check(locals(), "p2", (3,))
        vg.shape.check(locals(), "vector", (3,))

        normal = np.cross(p2 - p1, vector)

        return cls(point_on_plane=p1, unit_normal=normal)

    @classmethod
    def fit_from_points(cls, points):
        """
        Fits a plane whose normal is orthgonal to the first two principal axes
        of variation in the data and centered on their centroid.
        """
        vg.shape.check(locals(), "points", (-1, 3))

        eigval, eigvec = np.linalg.eig(np.cov(points.T))
        ordering = np.argsort(eigval)[::-1]
        normal = np.cross(eigvec[:, ordering[0]], eigvec[:, ordering[1]])

        centroid = points.mean(axis=0)

        return cls(centroid, normal)

    @property
    def equation(self):
        """
        Returns parameters `A`, `B`, `C`, `D` as a 1x4 `np.array`, where

            `Ax + By + Cz + D = 0`

        defines the plane.
        """
        A, B, C = self._n
        D = -self._r0.dot(self._n)

        return np.array([A, B, C, D])

    @property
    def reference_point(self):
        """
        The point used to create this plane.

        """
        return self._r0

    @property
    def canonical_point(self):
        """
        A canonical point on the plane, the one at which the normal
        would intersect the plane if drawn from the origin (0, 0, 0).

        This is computed by projecting the reference point onto the
        normal.

        This is useful for partitioning the space between two planes,
        as we do when searching for planar cross sections.

        """
        return self._r0.dot(self._n) * self._n

    @property
    def normal(self):
        """
        Return the plane's normal vector.

        """
        return self._n

    def flipped(self):
        """
        Creates a new Plane with an inverted orientation.
        """
        return Plane(point_on_plane=self._r0, unit_normal=-self._n)

    def sign(self, points):
        """
        Given an array of points, return an array with +1 for points in front
        of the plane (in the direction of the normal), -1 for points behind
        the plane (away from the normal), and 0 for points on the plane.

        """
        return np.sign(self.signed_distance(points))

    def points_in_front(self, points, inverted=False, ret_indices=False):
        """
        Given an array of points, return the points which lie either on the
        plane or in the half-space in front of it (i.e. in the direction of
        the plane normal).

        Args:
            points (np.arraylikw): An array of points.
            inverted (bool): When `True`, invert the logic. Return the points
                that lie behind the plane instead.
            ret_indices (bool): When `True`, return the indices instead of the
                points themselves.
        """
        sign = self.sign(points)

        if inverted:
            mask = np.less_equal(sign, 0)
        else:
            mask = np.greater_equal(sign, 0)

        indices = np.flatnonzero(mask)

        return indices if ret_indices else points[indices]

    def signed_distance(self, points):
        """
        Returns the signed distances to the given points or the signed
        distance to a single point.

        Args:
            points (np.arraylike): A 3D point or a `kx3` stack of points.

        Returns:
            depends:

            - Given a single 3D point, the distance as a NumPy scalar.
            - Given a `kx3` stack of points, an `k` array of distances.
        """
        return signed_distance_to_plane(points, self.equation)

    def distance(self, points):
        return np.absolute(self.signed_distance(points))

    def project_point(self, points):
        """
        Project a given point (or stack of points) to the plane.
        """
        return project_point_to_plane(points, self.equation)

    def line_xsection(self, pt, ray):
        vg.shape.check(locals(), "pt", (3,))
        vg.shape.check(locals(), "ray", (3,))
        return self._line_xsection(np.asarray(pt).ravel(), np.asarray(ray).ravel())

    def _line_xsection(self, pt, ray):
        denom = np.dot(ray, self.normal)
        if denom == 0:
            return None  # parallel, either coplanar or non-intersecting
        p = np.dot(self.reference_point - pt, self.normal) / denom
        return p * ray + pt

    def line_segment_xsection(self, a, b):
        vg.shape.check(locals(), "a", (3,))
        vg.shape.check(locals(), "b", (3,))
        return self._line_segment_xsection(np.asarray(a).ravel(), np.asarray(b).ravel())

    def _line_segment_xsection(self, a, b):
        pt = self._line_xsection(a, b - a)
        if pt is not None:
            if any(np.logical_and(pt > a, pt > b)) or any(
                np.logical_and(pt < a, pt < b)
            ):
                return None
        return pt

    def line_xsections(self, pts, rays):
        k = vg.shape.check(locals(), "pts", (-1, 3))
        vg.shape.check(locals(), "rays", (k, 3))
        denoms = np.dot(rays, self.normal)
        denom_is_zero = denoms == 0
        denoms[denom_is_zero] = np.nan
        p = np.dot(self.reference_point - pts, self.normal) / denoms
        return np.vstack([p, p, p]).T * rays + pts, ~denom_is_zero

    def line_segment_xsections(self, a, b):
        pts, pt_is_valid = self.line_xsections(a, b - a)
        pt_is_out_of_bounds = np.logical_or(
            np.any(
                np.logical_and(
                    pts[pt_is_valid] > a[pt_is_valid], pts[pt_is_valid] > b[pt_is_valid]
                ),
                axis=1,
            ),
            np.any(
                np.logical_and(
                    pts[pt_is_valid] < a[pt_is_valid], pts[pt_is_valid] < b[pt_is_valid]
                ),
                axis=1,
            ),
        )
        pt_is_valid[pt_is_valid] = ~pt_is_out_of_bounds
        pts[~pt_is_valid] = np.nan
        return pts, pt_is_valid


Plane.xy = Plane(point_on_plane=np.zeros(3), unit_normal=vg.basis.z)
Plane.xy.__doc__ = "The `xy`-plane."
Plane.xz = Plane(point_on_plane=np.zeros(3), unit_normal=vg.basis.y)
Plane.xz.__doc__ = "The `xz`-plane."
Plane.yz = Plane(point_on_plane=np.zeros(3), unit_normal=vg.basis.x)
Plane.yz.__doc__ = "The `yz`-plane."



import numpy as np
import vg


def quads_to_tris(quads, ret_mapping=False):
    """
    Convert quad faces to triangular faces.

    quads: An nx4 array.
    ret_mapping: A bool.

    When `ret_mapping` is `True`, return a 2nx3 array of new triangles and a 2nx3
    array mapping old quad indices to new trangle indices.

    When `ret_mapping` is `False`, return the 2nx3 array of triangles.
    """
    vg.shape.check(locals(), "quads", (-1, 4))

    tris = np.empty((2 * len(quads), 3))
    tris[0::2, :] = quads[:, [0, 1, 2]]
    tris[1::2, :] = quads[:, [0, 2, 3]]
    if ret_mapping:
        f_old_to_new = np.arange(len(tris)).reshape(-1, 2)
        return tris, f_old_to_new
    else:
        return tris



from setuptools import find_packages, setup

# Set version_info[__version__], while avoiding importing numpy, in case numpy
# and vg are being installed concurrently.
# https://packaging.python.org/guides/single-sourcing-package-version/
version_info = {}
exec(open("polliwog/package_version.py").read(), version_info)

with open("README.md") as f:
    readme = f.read()

with open("requirements.txt") as f:
    install_requires = f.read()

setup(
    name="polliwog",
    version=version_info["__version__"],
    description="2D and 3D computational geometry library which scales from prototyping to production",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Body Labs, Metabolize, and other contributors",
    author_email="github@paulmelnikow.com",
    url="https://github.com/lace/polliwog",
    project_urls={
        "Issue Tracker": "https://github.com/lace/polliwog/issues",
        "Documentation": "https://polliwog.readthedocs.io/en/stable/",
    },
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)



import vg
from ._composite_transform import CompositeTransform


class CoordinateManager(object):
    """
    Example:
        >>> coordinate_manager = CoordinateManager()
        >>> coordinate_manager.tag_as('source')
        >>> coordinate_manager.translate(-cube.floor_point)
        >>> coordinate_manager.scale(2)
        >>> coordinate_manager.tag_as('floored_and_scaled')
        >>> coordinate_manager.translate(np.array([0., -4., 0.]))
        >>> coordinate_manager.tag_as('centered_at_origin')

        >>> coordinate_manager.source = cube
        >>> centered_mesh = coordinate_manager.centered_at_origin
    """

    def __init__(self):
        self.__dict__.update(
            {
                # A map from tag names to indices into the transform stack.
                "_tags_to_indices": {},
                # Our currently set points, and the tag at which they belong.
                "_points_tag": None,
                "_points": None,
                # Our worthy collaborator.
                "_transform": CompositeTransform(),
            }
        )

    def append_transform(self, *args, **kwargs):
        self._transform.append_transform(*args, **kwargs)

    def scale(self, *args, **kwargs):
        self._transform.scale(*args, **kwargs)

    def convert_units(self, *args, **kwargs):
        self._transform.convert_units(*args, **kwargs)

    def translate(self, *args, **kwargs):
        self._transform.translate(*args, **kwargs)

    def reorient(self, *args, **kwargs):
        self._transform.reorient(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        self._transform.rotate(*args, **kwargs)

    def tag_as(self, name):
        """
        Give a name to the current state.

        """
        # The indices of CompositeTransform are 0-based, which means the first
        # transform is transform 0.
        #
        # In CoordinateManager, we refer to the initial state -- i.e. no
        # transforms -- with 0. The state after the first transform is 1.
        # After two transforms, 2. Put another way: we refer to a given state
        # with a value that is *one more than* the index of the state's last
        # transform. This is a little strange, but we need the extra zero
        # state. And it ends up playing nicely with array slicing.
        self._tags_to_indices[name] = len(self._transform.transforms)

    def do_transform(self, points, from_tag, to_tag):
        try:
            from_index = self._tags_to_indices[from_tag]
            to_index = self._tags_to_indices[to_tag]
        except KeyError as e:
            tag = e.args[0]
            raise KeyError("No such tag: {}".format(tag))

        if from_index == to_index:
            return points
        elif from_index < to_index:
            from_range = from_index, to_index
            return self._transform(points, from_range=from_range)
        else:
            from_range = to_index, from_index
            return self._transform(points, from_range=from_range, reverse=True)

    def __setattr__(self, name, points):
        """
        value: An nx3 array of points or an instance of Mesh.
        """
        if name not in self._tags_to_indices:
            raise AttributeError("No such tag: %s" % name)

        vg.shape.check(locals(), "points", (-1, 3))

        self.__dict__["_points_tag"] = name
        self.__dict__["_points"] = points

    def __getattr__(self, name):
        from_tag = self._points_tag
        if from_tag is None:
            raise ValueError("Must set the points before trying to read them")

        return self.do_transform(
            points=self._points, from_tag=self._points_tag, to_tag=name
        )



import numpy as np
import vg
from .._common.shape import columnize


def intersect_segment_with_plane(
    start_points, segment_vectors, points_on_plane, plane_normals
):
    """
    Check for intersections between a line segment and a plane, or pairwise
    between a stack of line segments and a stack of planes.
    """
    orig_shape = start_points.shape
    start_points, _, transform_result = columnize(
        start_points, (-1, 3), name="start_points"
    )
    vg.shape.check(locals(), "segment_vectors", orig_shape)
    vg.shape.check(locals(), "points_on_plane", orig_shape)
    vg.shape.check(locals(), "plane_normals", orig_shape)

    # Compute t values such that
    # `result = reference_point + t * segment_vectors`.
    t = np.nan_to_num(
        vg.dot(points_on_plane - start_points, plane_normals)
        / vg.dot(segment_vectors, plane_normals)
    )

    intersection_points = start_points + t.reshape(-1, 1) * segment_vectors

    # Discard points which lie past the ends of the segment.
    intersection_points[t < 0] = np.nan
    intersection_points[t > 1] = np.nan

    return transform_result(intersection_points)
